# heavily adpted from: https://raw.githubusercontent.com/PaoloLRinaldi/progress_bar_python/master/perc.py
from datetime import datetime, timedelta
from statistics import mean, stdev
import time
import sys
import math
from io import StringIO

from .__dependencies__.super_map import LazyDict

# nested indentation support
try:
    from blissful_basics import print as bliss_print
except Exception as error:
    # create a placeholder/stand-in
    def bliss_print(*args, **kwargs):
        print(*args, **kwargs)
    class Indent:
        size = 0
        string = "    "
    bliss_print.indent = Indent()

# GUI progress support in notebooks
try:
    from IPython.display import display, HTML, clear_output
    from io import StringIO
    ipython_exists = False
    
    try:
        from IPython.display import get_ipython
        ipython_exists = 'IPKernelApp' in get_ipython().config
        ipython_exists = True
    except Exception as error:
        pass
    
    try:
        from google.colab import output
        ipython_exists = True
    except Exception as error:
        pass
    
except ImportError:
    ipython_exists = False
except AttributeError:
    ipython_exists = False
    
class NotGiven: pass

def subsequence_replace(a_list, sequence, replacement):
    que = []
    new_list = []
    for each in a_list:
        que.append(each)
        while len(que) > len(sequence):
            new_list.append(que.pop(0))
        if que == sequence:
            que.clear()
            for each in replacement:
                new_list.append(each)
    new_list += que # add any remaining elements
    return new_list    

def to_time_string(secs):
    secs  = int(round(secs))
    mins  = secs  // 60; secs  = secs  % 60
    hours = mins  // 60; mins  = mins  % 60
    days  = hours // 24; hours = hours % 24
    if days:
        hours = f"{hours}".rjust(2,"0")
        mins  = f"{mins}".rjust(2,"0")
        return f"{days}days, {hours}h:{mins}min"
    elif hours:
        mins  = f"{mins}".rjust(2,"0")
        return f"{hours}h:{mins}min"
    elif mins:
        secs  = f"{secs}".rjust(2,"0")
        return f"{mins}m:{secs}sec"
    else:
        return f"{secs}sec"

nested_progress_bars = []
class ProgressBar:
    """
    Example:
        from informative_iterator import ProgressBar, time
        for progress, each in ProgressBar(10000):
            time.sleep(0.01)
    """
    
    layout = [ 'title', 'bar', 'percent', 'spacer', 'fraction', 'spacer', 'remaining_time', 'spacer', 'end_time', 'spacer', 'duration', 'spacer', ]
    minimal_layout = [ 'title', 'bar', 'spacer', 'end_time', 'spacer', ]
    spacer = " | "
    minmal = False
    inline = True
    disable_logging = False
    progress_bar_size = 35
    seconds_per_print = 0.1
    percent_per_print = 2
    lookback_size = 100
    time_format = "%H:%M:%S"
    long_time_format = "%D %H:%M:%S"
    
    @classmethod
    def configure(this_class, **config):
        for each_key, each_value in config.items():
            setattr(this_class, each_key, each_value)
    
    def __init__(self, iterator, *, title=None, iterations=None, layout=None, disable_logging=NotGiven, minimal=NotGiven, inline=NotGiven, progress_bar_size=NotGiven, seconds_per_print=NotGiven, percent_per_print=NotGiven, lookback_size=NotGiven):
        original_generator = range(int(iterator)) if isinstance(iterator, (int, float)) else iterator
        self.title = title or ""
        
        # inherit unspecified options from class object
        for each_option in [ "disable_logging", "minimal", "inline", "progress_bar_size", "seconds_per_print", "percent_per_print", "lookback_size" ]:
            arg_value = eval(each_option, locals())
            # default to the class value if not given
            if arg_value == NotGiven:
                actual_value = getattr(ProgressBar, each_option, None)
            # otherwise use the given value
            else:
                actual_value = arg_value
            # set the object's value
            setattr(self, each_option, actual_value)
        
        # initilize misc values
        self.past_indicies     = []
        self.start_time        = datetime.now()
        self.next_percent_mark = self.percent_per_print
        self.prev_time         = -math.inf
        self.times             = [time.time()]
        self.opacity = 0.7
        self.colors = LazyDict(
            progress="#9b68ab",
        )
        self.progress_data     = LazyDict(
            index=0,
            percent=0,
            updated=True,
            time=self.times[0],
            total_iterations=(len(original_generator) if iterations is None else iterations),
            deviation=None,
            expected_number_of_updates_needed=None,
            pretext="",
            text="",
            previous_output="",
        )
        # setup print
        if self.disable_logging:
            self.print = lambda *args, **kwargs: None
        elif not ipython_exists:
            self.print = print
        else:
            # remove the progress bar and percent
            layout = list(self.layout)
            layout = subsequence_replace(layout, [ 'spacer', 'bar'    , 'spacer' ], ['spacer'])
            layout = subsequence_replace(layout, [ 'spacer', 'bar'    ,          ], ['spacer'])
            layout = subsequence_replace(layout, [           'bar'    , 'spacer' ], ['spacer'])
            layout = subsequence_replace(layout, [           'bar'               ], [])
            layout = subsequence_replace(layout, [ 'spacer', 'percent', 'spacer' ], ['spacer'])
            layout = subsequence_replace(layout, [ 'spacer', 'percent',          ], ['spacer'])
            layout = subsequence_replace(layout, [           'percent', 'spacer' ], ['spacer'])
            layout = subsequence_replace(layout, [           'percent',          ], [])
            layout = subsequence_replace(layout, [ 'spacer', 'spacer'            ], ['spacer'])
            self.layout = layout
            self.string_buffer = ""
            self.html_created = False
            self.should_flush = False
            def ipython_print(*args, **kwargs):
                # get the string value
                string_stream = StringIO()
                print(*args, **kwargs, file=string_stream)
                output_str = string_stream.getvalue()
                string_stream.close()
                self.string_buffer += output_str
                # escape html just encase
                self.string_buffer = self.string_buffer.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace('‘', "'").replace('"', "&quot;").replace("\n", "<br>")
                
                # clear output whenever newline is created
                if self.should_flush:
                    self.should_flush = False
                    if not self.html_created:
                        self.html_created = True
                        clear_output(wait=True)
                        display(HTML(f'''
                            <div id="progressContainer" style="left: 0; top: 0; width: 95%; background: transparent; color: white; position: sticky; padding: 1.2rem 0.4rem; box-sizing: border-box;">
                                <div style="position: relative; background: #46505a; height: 2.7rem; width: 100%; border-radius: 10rem; border: transparent solid 0.34rem; box-sizing: border-box; opacity: {self.opacity};">
                                    <!-- color bar -->
                                    <div id="progressBar" style="height: 100%; background: {self.colors.progress}; border-radius: 10rem; transition: all 0.5s ease-in-out 0s;"></div>
                                    <!-- percentage text -->
                                    <div style="height: 100%; width: 100%; position: absolute; top: 0; left: 0; display: flex; flex-direction: column; align-items: center; align-content: center; justify-items: center;  justify-content: center;">
                                        <span id="progressPercent">
                                            {self.progress_data.percent:0.2f}%
                                        </span>
                                    </div>
                                </div>
                                <div style="width: 100%; height: 1rem;">
                                </div>
                                <div style="position: relative; height: fit-content; width: 100%; box-sizing: border-box; display: flex; flex-direction: column; align-items: center; align-content: center; justify-items: center;  justify-content: center;">
                                    <div style="position: relative;background: #46505a;height: fit-content;width: fit-content; min-width: 50%; border-radius: 1.2rem;border: transparent solid 0.5rem;box-sizing: border-box;display: flex;flex-direction: column;align-items: center;align-content: center;justify-items: center;justify-content: center;padding: 1rem; padding-top: 0;">
                                        <code id="progressText" style="white-space: pre; color: whitesmoke;" >{self.string_buffer}</code>
                                        <code id="customProgressText" style="white-space: pre; color: whitesmoke; padding-top: 1rem;" >{self.progress_data.text}</code>
                                    </div>
                                </div>
                                <style>
                                </style>
                                <script>
                                    var progressContainer = document.getElementById("progressContainer")
                                    var progressFooter = document.getElementById("progressFooter")
                                    var outputArea = document.getElementById("output-area")
                                    var outputHeader = document.getElementById("output-header")
                                    var outputBody = document.getElementById("output-body")
                                    outputHeader.appendChild(progressContainer)
                                    
                                    var scrollBoxSize = 27
                                    // outputArea
                                    outputArea.style.maxHeight = `${"{scrollBoxSize+10}"}rem`
                                    outputArea.style.minHeight = `${"{scrollBoxSize+10}"}rem`

                                    // outputHeader
                                    outputHeader.style.position = "absolute"
                                    outputHeader.style.top = "0"
                                    outputHeader.style.left = "0"
                                    outputHeader.style.width = "100%"

                                    // outputBody
                                    outputBody.style.paddingTop = "10.5rem"
                                    outputBody.style.position = "absolute"
                                    outputBody.style.height = `${"{scrollBoxSize}"}rem`
                                    outputBody.style.overflow = "auto"
                                    outputBody.style.minHeight = "50vh"
                                    outputBody.style.top = "0"
                                    outputBody.style.width = "97%"

                                    outputBody.scrollTo(0, outputBody.scrollHeight)
                                </script>
                            </div>
                        '''))
                    else:
                        from random import random
                        random_id_1 = f"id_{random()}".replace('.','')
                        random_id_2 = f"id_{random()}".replace('.','')
                        display(HTML(f'''
                            <div>
                                <code id="{random_id_1}" style="display: none;" >{self.string_buffer}</code>
                                <code id="{random_id_2}" style="display: none;" >{self.progress_data.text}</code>
                                <script>
                                    var bar = document.getElementById("progressBar")
                                    var percent = document.getElementById("progressPercent")
                                    var text = document.getElementById("progressText")
                                    var customText = document.getElementById("customProgressText")
                                    var textContainer = document.getElementById("{random_id_1}")
                                    var customTextContainer = document.getElementById("{random_id_2}")
                                    bar.style.width = `{self.progress_data.percent}%`
                                    percent.innerHTML = `{self.progress_data.percent:0.2f}%`
                                    // swap out contents (performed this way so that self.string_buffer uses html-escapes instead of javascript-string escapes)
                                    text.innerHTML = textContainer.innerHTML
                                    if (customTextContainer.innerHTML.trim().length > 0) customText.innerHTML = customTextContainer.innerHTML
                                </script>
                            </div>
                        '''))
                        
                    # clear the buffer
                    self.string_buffer = ""
            self.print = ipython_print
        
        # wrap it to record the text
        inner_print = self.print
        def print_capture(*args, **kwargs):
            string_stream = StringIO()
            output = print(*args, **kwargs, file=string_stream)
            inner_print(*args, **kwargs)
            output_str = string_stream.getvalue()
            self.progress_data.previous_output += output_str
            string_stream.close()
            return output
        self.print = print_capture
        
            
        # setup layout
        if layout == None and self.minimal:
            self.layout = ProgressBar.minimal_layout
        elif layout == None:
            self.layout = ProgressBar.layout
        else:
            self.layout = layout
        
        def generator_func():
            self.parent_bars = list(nested_progress_bars)
            nested_progress_bars.append(self)
            self.nested_indent = bliss_print.indent.string * len(self.parent_bars)
            if len(self.parent_bars) > 0:
                self.print()
            for self.progress_data.index, each_original in enumerate(original_generator):
                self.progress_data.previous_output = ""
                # update
                self.progress_data.time    = time.time()
                self.progress_data.updated = (self.progress_data.time - self.prev_time) > self.seconds_per_print
                self.progress_data.percent = (self.progress_data.index * 10000 // self.progress_data.total_iterations) / 100 # two decimals of accuracy
                
                if self.progress_data.updated:
                    self.prev_time = self.progress_data.time
                    
                # also printout at each percent marker
                if self.progress_data.percent >= self.next_percent_mark:
                    self.next_percent_mark += self.percent_per_print
                    self.progress_data.updated = True
                
                self.times.append(self.progress_data.time)
                self.past_indicies.append(self.progress_data.index)
                
                if self.progress_data.updated:
                    self.total_eslaped_time = 0
                    self.secs_remaining     = math.inf
                    # if more than 3, then stdev can be computed
                    if len(self.times) > 3:
                        self.total_eslaped_time = self.times[-1] - self.times[ 0]
                        
                        # 
                        # compute ETA as a slight overestimate that is less-of-an-overesitmate over time
                        # 
                        lookback_size = self.lookback_size
                        remaining_number_of_iterations = self.progress_data.total_iterations - self.progress_data.index
                        recent_indicies                = self.past_indicies[-lookback_size:]
                        recent_update_times            = self.times[-lookback_size:]
                        number_of_indicies_processed   =     recent_indicies[-1] -     recent_indicies[0]
                        time_span_of_recent_updates    = recent_update_times[-1] - recent_update_times[0]
                        time_per_update                = time_span_of_recent_updates / (len(recent_update_times)-1)
                        
                        list_of_iterations_per_update = tuple(each-prev for prev, each in zip(recent_indicies[0:-1], recent_indicies[1:]))
                        average_number_of_iterations_per_update = mean(list_of_iterations_per_update)
                        stdev_of_iters_per_update               = stdev(list_of_iterations_per_update) # TODO: the proper way to do this would be with a one sided bell curve
                        partial_deviation                       = (1 - (self.progress_data.percent/100)) * stdev_of_iters_per_update
                        iterations_per_update_lowerbound        = max((average_number_of_iterations_per_update-partial_deviation, min(list_of_iterations_per_update)))
                        soft_lowerbound                         = mean((iterations_per_update_lowerbound, average_number_of_iterations_per_update))
                        expected_number_of_updates_needed       = remaining_number_of_iterations / soft_lowerbound
                        
                        self.secs_remaining = time_per_update * expected_number_of_updates_needed
                    
                    # otherwise, use super basic linear prediction
                    elif len(self.times) > 2:
                        # provide very rough estimate on first iteration ()
                        self.total_eslaped_time = self.times[-1] - self.times[ 0]
                        self.secs_remaining = self.total_eslaped_time * (self.progress_data.total_iterations - 1)
                        
                        
                    indent = bliss_print.indent.string*bliss_print.indent.size + self.nested_indent
                    if self.progress_data.pretext:
                        self.print('', end='\r')
                        self.print('                                                                                                                        ', end='\r')
                        bliss_print(self.progress_data.pretext, end='\n')
                    
                    if self.inline:
                        self.print('', end='\r')
                    
                    self.print(indent, end="")
                    
                    # display each thing according to the layout
                    for each in self.layout:
                        getattr(self, f"show_{each}", lambda : None)()
                    
                    if not ipython_exists:
                        if self.progress_data.text:
                            line_1, *other_lines = self.progress_data.text.split('\n')
                            self.print(line_1, end='')
                            if other_lines:
                                nextline_text = f"\n{indent}" + f"\n{indent}".join(other_lines)
                                self.print(nextline_text, end='')
                        
                    if not self.inline:
                        self.print()
                    
                    # convoluted so that it handles both GUI and CLI
                    self.should_flush = True
                    self.print(end="")
                    sys.stdout.flush()
                    
                yield self.progress_data, each_original
                # manual stop if given an infinite generator and a "total_iterations" argument
                if self.progress_data.index+1 >= self.progress_data.total_iterations:
                    break
            
            self.progress_data.previous_output = ""
            self.show_done()
            nested_progress_bars.remove(self)
            
        self.iterator = iter(generator_func())

    def show_spacer(self):
        self.print(self.spacer, end='')
    
    def show_title(self):
        if self.title is not None:
            self.print(self.title, end=' ')
        
    def show_bar(self):
        prog = int((self.progress_data.index / self.progress_data.total_iterations) * self.progress_bar_size)
        self.print('[' + '=' * prog, end='')
        if prog != self.progress_bar_size:
            self.print('>' + '.' * (self.progress_bar_size - prog - 1), end='')
        self.print('] ', end='')
    
    def show_remaining_time(self):
        if self.secs_remaining == math.inf:
            self.print(f'remaining: _______', end='')
        elif self.progress_data.percent != 100:
            self.print(f'remaining: {to_time_string(self.secs_remaining)}', end='')
        
    def show_percent(self):
        self.print(f'{self.progress_data.percent:.2f}%'.rjust(6), end='')
    
    def show_duration(self):
        self.print("elapsed: "+to_time_string(self.total_eslaped_time), end='')
    
    def show_fraction(self):
        total_str = f"{self.progress_data.total_iterations}"
        self.print(f'{self.progress_data.index}'.rjust(len(total_str))+f'/{self.progress_data.total_iterations}', end='')
    
    def show_start_time(self):
        if self.progress_data.percent != 100:
            self.print(f'started: {self.start_time.strftime(self.time_format)}',  end='')
    
    def show_end_time(self):
        if self.progress_data.percent != 100:
            time_format = self.time_format
            if self.secs_remaining > (86400/2): # more than half a day
                time_format = self.long_time_format
            try:
                endtime = self.start_time + timedelta(seconds=self.total_eslaped_time + self.secs_remaining)
                self.print(f'eta: {endtime.strftime(time_format)}',  end='')
            except:
                self.print(f'eta: {"_"*(len(time_format)-3)}',  end='')
    
    def show_done(self):
        if self.inline:
            print("")
        duration = to_time_string(time.time() - self.times[0])
        end_time = datetime.now().strftime(self.time_format)
        self.progress_data.percent = 100.0
        self.string_buffer = "" # for ipython
        indent = (bliss_print.indent.string * bliss_print.indent.size) + self.nested_indent
        self.print(f'{indent}Done in {duration} at {end_time}')

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)
    
    def __len__(self):
        return self.progress_data.total_iterations



from warnings import warn
import pickle
import os
import math
import random
import functools

from .__dependencies__.ez_yaml import yaml
from .__dependencies__ import ez_yaml
from .__dependencies__.blissful_basics import FS, bytes_to_valid_string, valid_string_to_bytes, indent, super_hash, print, randomly_pick_from, stringify, to_pure
from .__dependencies__.informative_iterator import ProgressBar

# Version 1.0
    # DONE: add counting-caps (max IO for a particular function, or in-general)
    # add CLI tools
    # create add_input_for(func_id, args, kwargs, input_name)
    # use threads to offload the work
    # report which tests have recordings but were not tested during replay mode (e.g. couldn't reach/find function)
    # autodetect the git folder
# Version 2.0
    # fuzzing/coverage-tools; like analyzing boolean arguments, and generating inputs for all combinations of them
    # option to record stdout/stderr of a function
    # add `additional_inputs` in the decorator
    # add file path args to the decorator that create file copies, then inject/replace the path arguments

yaml.width = float("Infinity")

@yaml.register_class
class YamlPickled:
    yaml_tag = "!python/pickled"
    delimiter = ":"
    
    def __init__(self, value):
        self.value = value
    
    @classmethod
    def from_yaml(cls, constructor, node):
        string = node.value[node.value.index(YamlPickled.delimiter)+1:]
        # node.value is the python-value
        return pickle.loads(valid_string_to_bytes(string))
    
    @classmethod
    def to_yaml(cls, representer, data):
        prefix = f"{type(data.value)}".replace(YamlPickled.delimiter, "")
        if prefix.startswith("<class '") and prefix.endswith("'>"):
            prefix = prefix[8:-2]
            
        # value needs to be a string (or some other yaml-primitive)
        return representer.represent_scalar(
            tag=cls.yaml_tag,
            value=prefix + YamlPickled.delimiter + bytes_to_valid_string(
                pickle.dumps(data.value, protocol=4)
            ),
            style=None,
            anchor=None
        )
# 
# add yaml representations for numpy values if possible
# 
try:
    import numpy
    ez_yaml.yaml.Representer.add_representer(
        numpy.ndarray,
        lambda dumper, data: dumper.represent_sequence(tag='python/numpy/ndarray', sequence=data.tolist()), 
    )
    ez_yaml.ruamel.yaml.RoundTripConstructor.add_constructor(
        'python/numpy/ndarray',
        lambda loader, node: numpy.array(loader.construct_sequence(node, deep=True)),
    )
    
    # some types are commented out because I'm unsure about them loosing precision when being re-created and I didn't feel like testing to find out
    for each in [
        "float",
        'double',
        # "cfloat",
        # 'cdouble',
        'float8',
        'float16',
        'float32',
        'float64',
        # 'float128',
        # 'float256',
        # "longdouble",
        # "longfloat",
        # "clongdouble",
        # "clongfloat",
    ]:
        the_type = getattr(numpy, each, None)
        if the_type:
            the_tag = f'python/numpy/{each}'
            ez_yaml.yaml.Representer.add_representer(
                the_type,
                lambda dumper, data: dumper.represent_scalar(
                    tag=the_tag,
                    value=str(float(data)),
                    style=None,
                    anchor=None
                ),
            )
            ez_yaml.ruamel.yaml.RoundTripConstructor.add_constructor(
                the_tag,
                lambda loader, node: the_type(node.value),
            )

    for each in [
        # "intp",
        # "uintp",
        # "intc",
        # "uintc",
        # "longlong",
        # "ulonglong",
        "int",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "uint128",
        "uint256",
        "int",
        "int8",
        "int16",
        "int32",
        "int64",
        "int128",
        "int256",
    ]:
        the_type = getattr(numpy, each, None)
        if the_type:
            the_tag = f'python/numpy/{each}'
            ez_yaml.yaml.Representer.add_representer(
                the_type,
                lambda dumper, data: dumper.represent_scalar(
                    tag=the_tag,
                    value=str(int(data)),
                    style=None,
                    anchor=None
                ),
            )
            ez_yaml.ruamel.yaml.RoundTripConstructor.add_constructor(
                the_tag,
                lambda loader, node: the_type(node.value),
            )
except Exception as error:
    pass

class GrugTest:
    """
        Example:
            grug_test = GrugTest(
                project_folder=".",
                test_folder="./tests/grug_tests",
                fully_disable=os.environ.get("PROD")!=None,
                record_io=os.environ.get("DEV")!=None,
            )
            
            @grug_test
            def add_nums(a,b):
                return a + b + 1

            # normal usage
            for a,b in zip(range(10), range(30, 40)):
                add_nums(a,b)
    """
    overflow_strats = [ 'keep_old', 'delete_random', ]
    input_file_extension = ".input.yaml"
    output_file_extension = ".output.yaml"
    def __init__(
        self,
        name="__default__",
        fully_disable=False,
        replay_inputs=False,
        record_io=True,
        verbose=False,
        max_io_per_func=None,
        overflow_strat="keep_old",
        project_folder="./", # FIXME: walk up until .git
        test_folder="./tests/grug_tests/", # FIXME: walk up until .git
        # use_threading=True,
    ):
        self.name            = name
        self.fully_disable   = fully_disable
        self.replay_inputs   = replay_inputs
        self.record_io       = record_io
        self.verbose         = verbose
        self.overflow_strat  = overflow_strat
        self.max_io_per_func = max_io_per_func if max_io_per_func != None else math.inf
        self.project_folder  = project_folder
        self.test_folder     = test_folder
        
        # 
        # setup grug info
        # 
        grug_info_path = f"{test_folder}/{self.name}.grug_info.yaml"
        self.grug_info = {}
        grug_info_string = FS.read(grug_info_path)
        if grug_info_string:
            self.grug_info = ez_yaml.to_object(string=grug_info_string)
            if type(self.grug_info) != dict:
                self.grug_info = {}
                warn(f"The grug_info at {repr(grug_info_path)} seems to be corrupted, it will be ignored")
        
        self.grug_info = {
            **self.grug_info,
            "functions_with_tests": self.grug_info.get("functions_with_tests", []),
        }
        self.has_been_tested = { name: False for name in self.grug_info["functions_with_tests"] }
    
    # 
    # decorator
    # 
    def __call__(self, *args, func_name=None, max_io=None, record_io=None, additional_io_per_run=None, **kwargs):
        """
        Example:
            grug_test = GrugTest(
                project_folder=".",
                test_folder="./tests/grug_tests",
                fully_disable=os.environ.get("PROD")!=None,
                record_io=os.environ.get("DEV")!=None,
            )
            
            @grug_test
            def add_nums(a,b):
                return a + b + 1

            # normal usage
            for a,b in zip(range(10), range(30, 40)):
                add_nums(a,b)
        
        """
        if record_io == None:
            record_io = self.record_io
        if max_io == None:
            max_io = self.max_io_per_func
        if additional_io_per_run == None:
            additional_io_per_run = math.inf
        
        source = _get_path_of_caller()
        def decorator(function_being_wrapped):
            nonlocal max_io
            if self.fully_disable:
                # no wrapping
                return function_being_wrapped
            
            function_name = None
            grug_folder_for_this_func = None
            decorator.record_io = record_io
            decorator.replaying_inputs = False
            if self.record_io or self.replay_inputs:
                # 
                # setup name/folder
                # 
                relative_path_to_function = FS.normalize(FS.make_relative_path(coming_from=self.project_folder, to=source))
                relative_path_to_test = self.test_folder+"/"+relative_path_to_function
                function_name = func_name or getattr(function_being_wrapped, "__name__", "<unknown_func>")
                function_id = f"{relative_path_to_function}:{function_name}"
                grug_folder_for_this_func = relative_path_to_test+"/"+function_name
                if function_id not in self.grug_info["functions_with_tests"]:
                    self.grug_info["functions_with_tests"].append(function_id)
                
                FS.ensure_is_folder(grug_folder_for_this_func)
                input_files = [ each for each in FS.list_file_paths_in(grug_folder_for_this_func) if each.endswith(self.input_file_extension) ]
                # convert additional_io_per_run to a max_io value
                if additional_io_per_run > 0 and max_io > len(input_files):
                    max_io = min(max_io, len(input_files) + additional_io_per_run)
                
                # 
                # replay inputs
                # 
                if self.replay_inputs and not self.has_been_tested.get(function_id, False):
                    if self.verbose: print(f"replaying inputs for: {function_name}")
                    decorator.replaying_inputs = True
                    original_record_io_value = decorator.record_io
                    for progress, path in ProgressBar(input_files, disable_logging=not self.verbose):
                        progress.text = f" loading: {FS.basename(path)}"
                        try:
                            args, kwargs = ez_yaml.to_object(file_path=path)["pickled_args_and_kwargs"]
                            output, the_error = self.record_output(
                                function_being_wrapped,
                                args,
                                kwargs,
                                path=path[0:-len(self.input_file_extension)] + self.output_file_extension,
                                function_name=function_name,
                                source=source,
                            )
                        except Exception as error:
                            warn(error)
                    decorator.replaying_inputs = False
                    self.has_been_tested[function_id] = True
            
            @functools.wraps(function_being_wrapped) # fixes the stack-trace to make the decorator invisible
            def wrapper(*args, **kwargs):
                # normal run
                if decorator.replaying_inputs or not decorator.record_io:
                    return function_being_wrapped(*args, **kwargs)
                
                # when overflowing, with 'keep_old' just avoid saving io (even though technically we might want to update the output of an existing one)
                is_overflowing = len(input_files) >= max_io
                shouldnt_save_new_io = is_overflowing and self.overflow_strat == 'keep_old'
                if shouldnt_save_new_io:
                    return function_being_wrapped(*args, **kwargs)
                
                # 
                # hash the inputs
                #
                input_hash = None
                try:
                    arg = (args, kwargs)
                    input_hash = super_hash(arg)[0:12] # 12 chars is plenty for being unique 
                except Exception as error:
                    error_message = f"\n\n\nFor a grug test on this function: {repr(function_name)}\n" + indent(f"I tried to hash the inputs but I wasn't able to.\nHere are the input types:\n    args: {indent(stringify(tuple(type(each) for each in args)))}\n    kwargs: {indent(stringify({ key: type(value) for key, value in kwargs.items()}))}\n\nAnd here's the error:\n{indent(error)}\n")
                    warn(error_message, category=None, stacklevel=1, source=source)
                    # run function like normal
                    return function_being_wrapped(*args, **kwargs)
                
                input_file_path  = grug_folder_for_this_func+f"/{input_hash}{self.input_file_extension}"
                output_file_path = grug_folder_for_this_func+f"/{input_hash}{self.output_file_extension}"
                
                try:
                    # 
                    # input limiter
                    # 
                    input_already_existed = FS.is_file(input_file_path)
                    if is_overflowing and not input_already_existed and self.overflow_strat == 'delete_random':
                        input_to_delete  = randomly_pick_from(input_files)
                        output_to_delete = input_to_delete[0:-len(self.input_file_extension)]+self.output_file_extension
                        FS.remove(input_to_delete)
                        FS.remove(output_to_delete)
                        input_files.remove(input_to_delete)
                    
                    # 
                    # save the inputs
                    # 
                    if not input_already_existed:
                        FS.ensure_is_folder(FS.parent_path(input_file_path))
                        # encase its a folder for some reason
                        FS.remove(input_file_path)
                        # if all the args are yaml-able this will work
                        try:
                            ez_yaml.to_file(
                                file_path=input_file_path,
                                obj=dict(
                                    args=args,
                                    kwargs=kwargs,
                                    pickled_args_and_kwargs=YamlPickled(arg),
                                ),
                            )
                        except Exception as error:
                            # if all the args are at least pickle-able, this will work
                            converted_args = list(args)
                            converted_kwargs = dict(kwargs)
                            for index,each in enumerate(converted_args):
                                try:
                                    ez_yaml.to_string(each)
                                except Exception as error:
                                    converted_args[index] = YamlPickled(each)
                            for each_key, each_value in converted_kwargs.items():
                                try:
                                    ez_yaml.to_string(each_value)
                                except Exception as error:
                                    converted_kwargs[each_key] = YamlPickled(each_value)
                            ez_yaml.to_file(
                                file_path=input_file_path,
                                obj=dict(
                                    args=converted_args,
                                    kwargs=converted_kwargs,
                                    pickled_args_and_kwargs=YamlPickled(arg),
                                )
                            )
                        input_files.append(input_file_path)
                except Exception as error:
                    FS.remove(input_file_path)
                    warn(f"\n\n\nFor a grug test on this function: {repr(function_name)}\n"+indent(f"I tried to seralize the inputs but I wasn't able to.\nHere are the input types:\n    args: {indent(stringify(tuple(type(each) for each in args)))}\n    kwargs: {indent(stringify({ key: type(value) for key, value in kwargs.items()}))}\n\nAnd here's the error:\n{indent(error)}\n"), category=None, stacklevel=1, source=source)
                    # run function like normal
                    return function_being_wrapped(*args, **kwargs)
                
                # 
                # save the output
                # 
                output, the_error = self.record_output(
                    function_being_wrapped,
                    args,
                    kwargs,
                    path=output_file_path,
                    function_name=function_name,
                    source=source,
                )
                
                # raise errors like normal
                if the_error != None:
                    raise the_error
                
                return output
            return wrapper
        
        # this handles
        # @grug_test
        # def thign(): pass
        if len(args) == 1 and callable(args[0]):
            return decorator(args[0])
        # this handles
        # @grug_test(options=somethin)
        # def thign(): pass
        else:
            return decorator
    
    @staticmethod
    def record_output(func, args, kwargs, path, function_name, source):
        the_error = None
        output = None
        try:
            output = func(*args, **kwargs)
        except Exception as error:
            the_error = error
        
        # clear the way (generates parent folders if needed)
        FS.ensure_is_folder(FS.parent_path(path))
        # encase its a folder for some reason
        FS.remove(path)
        try:
            ez_yaml.to_file(
                file_path=path,
                obj={
                    "error_output": repr(the_error),
                    "normal_output": output,
                },
            )
        except Exception as error:
            try:
                # try to be informative if possible
                if type(output) == tuple:
                    new_output = list(output)
                    for index, each in enumerate(output):
                        try:
                            ez_yaml.to_string(each)
                        except Exception as error:
                            new_output[index] = YamlPickled(each)
                                    
                    ez_yaml.to_file(
                        file_path=path,
                        obj={
                            "error_output": repr(the_error),
                            "normal_output": new_output,
                        },
                    )
                else:
                    ez_yaml.to_file(
                        file_path=path,
                        obj={
                            "error_output": repr(the_error),
                            "normal_output": YamlPickled(output),
                        },
                    )
            except Exception as error:
                FS.remove(path)
                warn(f"\n\n\nFor a grug test on this function: {repr(function_name)}\n"+ indent(f"I tried to seralize the output but I wasn't able to.\nHere is the output type:\n    output: {type(output)}\nAnd here's the error: {indent(error)}\n"), category=None, stacklevel=1, source=source)
    
        return output, the_error
        
def _get_path_of_caller(*paths):
    import os
    import inspect
    
    intial_cwd = os.getcwd()
    # https://stackoverflow.com/questions/28021472/get-relative-path-of-caller-in-python
    try:
        frame = inspect.stack()[2]
        module = inspect.getmodule(frame[0])
        directory = module.__file__
    # if inside a repl (error =>) assume that the working directory is the path
    except (AttributeError, IndexError) as error:
        directory = cwd
    
    if FS.is_absolute_path(directory):
        return FS.join(directory, *paths)
    else:
        # See note at the top
        return FS.join(intial_cwd, directory, *paths)


# coding: utf-8

from roundtrip import dedent


class TestDedent:
    def test_start_newline(self):
        # fmt: off
        x = dedent("""
        123
          456
        """)
        # fmt: on
        assert x == '123\n  456\n'

    def test_start_space_newline(self):
        # special construct to prevent stripping of following whitespace
        # fmt: off
        x = dedent("   " """
        123
        """)
        # fmt: on
        assert x == '123\n'

    def test_start_no_newline(self):
        # special construct to prevent stripping of following whitespac
        x = dedent("""\
        123
          456
        """)
        assert x == '123\n  456\n'

    def test_preserve_no_newline_at_end(self):
        x = dedent("""
        123""")
        assert x == '123'

    def test_preserve_no_newline_at_all(self):
        x = dedent("""\
        123""")
        assert x == '123'

    def test_multiple_dedent(self):
        x = dedent(
            dedent("""
        123
        """)
        )
        assert x == '123\n'



from grug_test import GrugTest
import os

grug_test = GrugTest(
    project_folder=".",
    test_folder="./tests/grug_tests",
    fully_disable=os.environ.get("PROD")!=None,
    record_io=os.environ.get("DEV")!=None,
)

@grug_test
def add_nums(a,b):
    return a + b + 1

# normal usage
for a,b in zip(range(10), range(30, 40)):
    add_nums(a,b)


# coding: utf-8

import pytest  # NOQA

from roundtrip import round_trip, dedent, round_trip_load, round_trip_dump  # NOQA


def load(s):
    return round_trip_load(dedent(s))


def compare(data, s, **kw):
    assert round_trip_dump(data, **kw) == dedent(s)


def compare_eol(data, s):
    assert 'EOL' in s
    ds = dedent(s).replace('EOL', '').replace('\n', '|\n')
    assert round_trip_dump(data).replace('\n', '|\n') == ds


class TestCommentsManipulation:

    # list
    def test_seq_set_comment_on_existing_explicit_column(self):
        data = load("""
        - a   # comment 1
        - b
        - c
        """)
        data.yaml_add_eol_comment('comment 2', key=1, column=6)
        exp = """
        - a   # comment 1
        - b   # comment 2
        - c
        """
        compare(data, exp)

    def test_seq_overwrite_comment_on_existing_explicit_column(self):
        data = load("""
        - a   # comment 1
        - b
        - c
        """)
        data.yaml_add_eol_comment('comment 2', key=0, column=6)
        exp = """
        - a   # comment 2
        - b
        - c
        """
        compare(data, exp)

    def test_seq_first_comment_explicit_column(self):
        data = load("""
        - a
        - b
        - c
        """)
        data.yaml_add_eol_comment('comment 1', key=1, column=6)
        exp = """
        - a
        - b   # comment 1
        - c
        """
        compare(data, exp)

    def test_seq_set_comment_on_existing_column_prev(self):
        data = load("""
        - a   # comment 1
        - b
        - c
        - d     # comment 3
        """)
        data.yaml_add_eol_comment('comment 2', key=1)
        exp = """
        - a   # comment 1
        - b   # comment 2
        - c
        - d     # comment 3
        """
        compare(data, exp)

    def test_seq_set_comment_on_existing_column_next(self):
        data = load("""
        - a   # comment 1
        - b
        - c
        - d     # comment 3
        """)
        print(data._yaml_comment)
        # print(type(data._yaml_comment._items[0][0].start_mark))
        # ruamel.yaml.error.Mark
        # print(type(data._yaml_comment._items[0][0].start_mark))
        data.yaml_add_eol_comment('comment 2', key=2)
        exp = """
        - a   # comment 1
        - b
        - c     # comment 2
        - d     # comment 3
        """
        compare(data, exp)

    def test_seq_set_comment_on_existing_column_further_away(self):
        """
        no comment line before or after, take the latest before
        the new position
        """
        data = load("""
        - a   # comment 1
        - b
        - c
        - d
        - e
        - f     # comment 3
        """)
        print(data._yaml_comment)
        # print(type(data._yaml_comment._items[0][0].start_mark))
        # ruamel.yaml.error.Mark
        # print(type(data._yaml_comment._items[0][0].start_mark))
        data.yaml_add_eol_comment('comment 2', key=3)
        exp = """
        - a   # comment 1
        - b
        - c
        - d   # comment 2
        - e
        - f     # comment 3
        """
        compare(data, exp)

    def test_seq_set_comment_on_existing_explicit_column_with_hash(self):
        data = load("""
        - a   # comment 1
        - b
        - c
        """)
        data.yaml_add_eol_comment('#  comment 2', key=1, column=6)
        exp = """
        - a   # comment 1
        - b   #  comment 2
        - c
        """
        compare(data, exp)

    # dict

    def test_dict_set_comment_on_existing_explicit_column(self):
        data = load("""
        a: 1   # comment 1
        b: 2
        c: 3
        d: 4
        e: 5
        """)
        data.yaml_add_eol_comment('comment 2', key='c', column=7)
        exp = """
        a: 1   # comment 1
        b: 2
        c: 3   # comment 2
        d: 4
        e: 5
        """
        compare(data, exp)

    def test_dict_overwrite_comment_on_existing_explicit_column(self):
        data = load("""
        a: 1   # comment 1
        b: 2
        c: 3
        d: 4
        e: 5
        """)
        data.yaml_add_eol_comment('comment 2', key='a', column=7)
        exp = """
        a: 1   # comment 2
        b: 2
        c: 3
        d: 4
        e: 5
        """
        compare(data, exp)

    def test_map_set_comment_on_existing_column_prev(self):
        data = load("""
            a: 1   # comment 1
            b: 2
            c: 3
            d: 4
            e: 5     # comment 3
            """)
        data.yaml_add_eol_comment('comment 2', key='b')
        exp = """
            a: 1   # comment 1
            b: 2   # comment 2
            c: 3
            d: 4
            e: 5     # comment 3
            """
        compare(data, exp)

    def test_map_set_comment_on_existing_column_next(self):
        data = load("""
            a: 1   # comment 1
            b: 2
            c: 3
            d: 4
            e: 5     # comment 3
            """)
        data.yaml_add_eol_comment('comment 2', key='d')
        exp = """
            a: 1   # comment 1
            b: 2
            c: 3
            d: 4     # comment 2
            e: 5     # comment 3
            """
        compare(data, exp)

    def test_map_set_comment_on_existing_column_further_away(self):
        """
        no comment line before or after, take the latest before
        the new position
        """
        data = load("""
            a: 1   # comment 1
            b: 2
            c: 3
            d: 4
            e: 5     # comment 3
            """)
        data.yaml_add_eol_comment('comment 2', key='c')
        print(round_trip_dump(data))
        exp = """
            a: 1   # comment 1
            b: 2
            c: 3   # comment 2
            d: 4
            e: 5     # comment 3
            """
        compare(data, exp)

    def test_before_top_map_rt(self):
        data = load("""
        a: 1
        b: 2
        """)
        data.yaml_set_start_comment('Hello\nWorld\n')
        exp = """
        # Hello
        # World
        a: 1
        b: 2
        """
        compare(data, exp.format(comment='#'))

    def test_before_top_map_replace(self):
        data = load("""
        # abc
        # def
        a: 1 # 1
        b: 2
        """)
        data.yaml_set_start_comment('Hello\nWorld\n')
        exp = """
        # Hello
        # World
        a: 1 # 1
        b: 2
        """
        compare(data, exp.format(comment='#'))

    def test_before_top_map_from_scratch(self):
        from .comments import CommentedMap

        data = CommentedMap()
        data['a'] = 1
        data['b'] = 2
        data.yaml_set_start_comment('Hello\nWorld\n')
        # print(data.ca)
        # print(data.ca._items)
        exp = """
            # Hello
            # World
            a: 1
            b: 2
            """
        compare(data, exp.format(comment='#'))

    def test_before_top_seq_rt(self):
        data = load("""
        - a
        - b
        """)
        data.yaml_set_start_comment('Hello\nWorld\n')
        print(round_trip_dump(data))
        exp = """
        # Hello
        # World
        - a
        - b
        """
        compare(data, exp)

    def test_before_top_seq_rt_replace(self):
        s = """
        # this
        # that
        - a
        - b
        """
        data = load(s.format(comment='#'))
        data.yaml_set_start_comment('Hello\nWorld\n')
        print(round_trip_dump(data))
        exp = """
        # Hello
        # World
        - a
        - b
        """
        compare(data, exp.format(comment='#'))

    def test_before_top_seq_from_scratch(self):
        from .comments import CommentedSeq

        data = CommentedSeq()
        data.append('a')
        data.append('b')
        data.yaml_set_start_comment('Hello\nWorld\n')
        print(round_trip_dump(data))
        exp = """
        # Hello
        # World
        - a
        - b
        """
        compare(data, exp.format(comment='#'))

    # nested variants
    def test_before_nested_map_rt(self):
        data = load("""
        a: 1
        b:
          c: 2
          d: 3
        """)
        data['b'].yaml_set_start_comment('Hello\nWorld\n')
        exp = """
        a: 1
        b:
        # Hello
        # World
          c: 2
          d: 3
        """
        compare(data, exp.format(comment='#'))

    def test_before_nested_map_rt_indent(self):
        data = load("""
        a: 1
        b:
          c: 2
          d: 3
        """)
        data['b'].yaml_set_start_comment('Hello\nWorld\n', indent=2)
        exp = """
        a: 1
        b:
          # Hello
          # World
          c: 2
          d: 3
        """
        compare(data, exp.format(comment='#'))
        print(data['b'].ca)

    def test_before_nested_map_from_scratch(self):
        from .comments import CommentedMap

        data = CommentedMap()
        datab = CommentedMap()
        data['a'] = 1
        data['b'] = datab
        datab['c'] = 2
        datab['d'] = 3
        data['b'].yaml_set_start_comment('Hello\nWorld\n')
        exp = """
        a: 1
        b:
        # Hello
        # World
          c: 2
          d: 3
        """
        compare(data, exp.format(comment='#'))

    def test_before_nested_seq_from_scratch(self):
        from .comments import CommentedMap, CommentedSeq

        data = CommentedMap()
        datab = CommentedSeq()
        data['a'] = 1
        data['b'] = datab
        datab.append('c')
        datab.append('d')
        data['b'].yaml_set_start_comment('Hello\nWorld\n', indent=2)
        exp = """
        a: 1
        b:
          # Hello
          # World
        - c
        - d
        """
        compare(data, exp.format(comment='#'))

    def test_before_nested_seq_from_scratch_block_seq_indent(self):
        from .comments import CommentedMap, CommentedSeq

        data = CommentedMap()
        datab = CommentedSeq()
        data['a'] = 1
        data['b'] = datab
        datab.append('c')
        datab.append('d')
        data['b'].yaml_set_start_comment('Hello\nWorld\n', indent=2)
        exp = """
        a: 1
        b:
          # Hello
          # World
          - c
          - d
        """
        compare(data, exp.format(comment='#'), indent=4, block_seq_indent=2)

    def test_map_set_comment_before_and_after_non_first_key_00(self):
        # http://stackoverflow.com/a/40705671/1307905
        data = load("""
        xyz:
          a: 1    # comment 1
          b: 2

        test1:
          test2:
            test3: 3
                """)
        data.yaml_set_comment_before_after_key(
            'test1', 'before test1 (top level)', after='before test2'
        )
        data['test1']['test2'].yaml_set_start_comment('after test2', indent=4)
        exp = """
        xyz:
          a: 1    # comment 1
          b: 2

        # before test1 (top level)
        test1:
          # before test2
          test2:
            # after test2
            test3: 3
        """
        compare(data, exp)

    def Xtest_map_set_comment_before_and_after_non_first_key_01(self):
        data = load("""
        xyz:
          a: 1    # comment 1
          b: 2

        test1:
          test2:
            test3: 3
        """)
        data.yaml_set_comment_before_after_key(
            'test1', 'before test1 (top level)', after='before test2\n\n'
        )
        data['test1']['test2'].yaml_set_start_comment('after test2', indent=4)
        # EOL is needed here as dedenting gets rid of spaces (as well as does Emacs
        exp = """
        xyz:
          a: 1    # comment 1
          b: 2

        # before test1 (top level)
        test1:
          # before test2
          EOL
          test2:
            # after test2
            test3: 3
        """
        compare_eol(data, exp)

    # EOL is no longer necessary
    # fixed together with issue # 216
    def test_map_set_comment_before_and_after_non_first_key_01(self):
        data = load("""
        xyz:
          a: 1    # comment 1
          b: 2

        test1:
          test2:
            test3: 3
        """)
        data.yaml_set_comment_before_after_key(
            'test1', 'before test1 (top level)', after='before test2\n\n'
        )
        data['test1']['test2'].yaml_set_start_comment('after test2', indent=4)
        exp = """
        xyz:
          a: 1    # comment 1
          b: 2

        # before test1 (top level)
        test1:
          # before test2

          test2:
            # after test2
            test3: 3
        """
        compare(data, exp)

    def Xtest_map_set_comment_before_and_after_non_first_key_02(self):
        data = load("""
        xyz:
          a: 1    # comment 1
          b: 2

        test1:
          test2:
            test3: 3
        """)
        data.yaml_set_comment_before_after_key(
            'test1', 'xyz\n\nbefore test1 (top level)', after='\nbefore test2', after_indent=4
        )
        data['test1']['test2'].yaml_set_start_comment('after test2', indent=4)
        # EOL is needed here as dedenting gets rid of spaces (as well as does Emacs
        exp = """
        xyz:
          a: 1    # comment 1
          b: 2

        # xyz

        # before test1 (top level)
        test1:
            EOL
            # before test2
          test2:
            # after test2
            test3: 3
        """
        compare_eol(data, exp)

    def test_map_set_comment_before_and_after_non_first_key_02(self):
        data = load("""
        xyz:
          a: 1    # comment 1
          b: 2

        test1:
          test2:
            test3: 3
        """)
        data.yaml_set_comment_before_after_key(
            'test1', 'xyz\n\nbefore test1 (top level)', after='\nbefore test2', after_indent=4
        )
        data['test1']['test2'].yaml_set_start_comment('after test2', indent=4)
        exp = """
        xyz:
          a: 1    # comment 1
          b: 2

        # xyz

        # before test1 (top level)
        test1:

            # before test2
          test2:
            # after test2
            test3: 3
        """
        compare(data, exp)
